==> v0.2.tar.gz: mailinglist.rb/lib/protocols/mail.rb <==
require "uuid"

class Protocols::Mail
  attr_accessor :from_name, :from, :to_name, :to, :subject, :body

  HEADERS_KEEP = %w[Message-ID Subject From To Date In-Reply-To References Content-Type User-Agent MIME-Version]
  USER_AGENT = "Ruby/#{RUBY_VERSION} (#{RUBY_PLATFORM}) Mailinglist.rb/1.0"

  def initialize(imap_mail: nil)
    if imap_mail
      @imap_mail = imap_mail
      envelope = imap_mail.attr[Protocols::ENVELOPE]
      body_head = imap_mail.attr[Protocols::HEADERS]
      body_text = imap_mail.attr[Protocols::BODYTEXT]
      @from_name = envelope.from.first.name
      @to_name   = envelope.to.first.name
      @from      = "#{envelope.from.first.mailbox}@#{envelope.from.first.host}"
      @to        = "#{envelope.to.first.mailbox}@#{envelope.to.first.host}"
      @subject   = envelope.subject
      @body      = body_text
      @headers   = parse_rfc822_headers!(body_head)
      @seq       = header("X-Sequence").to_i
      @message_id = header("Message-id")
    end
  end

  def clone
    Protocols::Mail.new(imap_mail: @imap_mail)
  end

  def seen!(imap_client:)
    if @imap_mail
      imap_client.seen!(@imap_mail)
    else
      false
    end
  end

  # Transform an received email into a mailing list email.
  # Tested with evolution. Thunderbird bug at the moment.
  #
  # @param list   [Mailinglist]
  # @param dest [Email]
  def to_redistribute(list:, dest:)
    new = clone
    new.replace_headers!(
      ["User-Agent",       USER_AGENT],
      # require people to do not respond privatly
      ["Reply-to",         "#{list.name} <#{list.email}>"],
      ["Errors-To",        list.email],
      ["To",               dest.email],
      ["From",             "\"#{@from_name}\" (via #{list.name}) <#{@from}>"],
      ["Sender",           @from],
      ["Date",             Time.now.to_s],
      ["List-Id",          "<#{list.email}>"],
      ["List-Post",        "<mailto:#{list.email}>"],
      ["List-Help",        "<mailto:#{list.help_email}>"],
      ["List-Subscribe",   "<mailto:#{list.subscribe_email}>"],
      ["List-Unsubscribe", "<mailto:#{list.unsubscribe_email}>"],
      ["List-Post",        "<mailto:#{list.post_email}>"],
      ["List-Owner",       "<mailto:#{list.owner_email}>"],
      ["In-Reply-To",      @message_id],
      ["Precedence",       "list"],
      # ["Precedence",       "bulk"],
      ["Message-Id",       "<#{UUID.generate}@#{ENV['SENDER_HOST']}>"],
      ["X-Sequence",         (@seq + 1).to_s],
      ["X-Loop",           list.email],
      ["Return-Path",      "<#{list.email}>"],
    )

    new.to   = dest.email
    new.from = list.email
    new.body = "#{body}#{list.signature}"

    new
  end

  # Create a email from scratch.
  #
  # @param list [Mailinglist]
  # @param to   [String]        email of the target
  # @param cc   [Array(String)] optional, emails (["mail@host", "Name <mail@host>", ...])
  def self.build(subject:, list:, to:, body:, cc: [])
    new = new()
    new.replace_headers!(
      ["User-Agent",       USER_AGENT],
      # require people to do not respond privatly
      ["Reply-to",         "#{list.name} <#{list.email}>"],
      ["Errors-To",        list.email],
      ["To",               to],
      ["Subject",          subject],
      ["From",             "\"#{list.name}\" <#{list.email}>"],
      ["Sender",           list.email],
      ["Date",             Time.now.to_s],
      ["Message-Id",       "<#{UUID.generate}@#{ENV['SENDER_HOST']}>"],
    )
    new.replace_headers!(["CC", cc.join(", ")]) if !cc.empty?

    new.subject = subject
    new.to   = to
    new.from = list.email
    new.body = "#{body}#{list.signature}"
    new
  end

  # Create a email from scratch.
  # Create a response email for an existing one.
  #
  # @param list [Mailinglist]
  # @param to   [Protocols::Mail]
  def to_response(list:, to:, body:)
    new = clone
    new.replace_headers!(
      ["User-Agent",       USER_AGENT],
      # require people to do not respond privatly
      ["Reply-to",         "#{list.name} <#{list.email}>"],
      ["Errors-To",        list.email],
      ["To",               list.email],
      ["From",             "\"#{list.name}\" <#{list.email}>"],
      ["Sender",           list.email],
      ["Date",             Time.now.to_s],
      ["Message-Id",       "<#{UUID.generate}@#{ENV['SENDER_HOST']}>"],
    )

    new.to   = to.from
    new.from = list.email
    new.body = "#{body}#{list.signature}"
    new
  end

  def parse_rfc822_headers!(rfc822)
    headers = []
    rfc822.split("\r\n").each do |line|
      if line[0] == "\t"
        headers.last[1] << "\r\n" << line
      else
        k, v = line.split(":", 2)
        headers << [k, v.to_s.strip]
      end
    end
    headers.filter! { HEADERS_KEEP.include?(_1[0]) }
  end

  def header(k)
    tuple = headers.find { _1[0].downcase == k.downcase }
    tuple && tuple[1]
  end

  def headers
    @headers ||= []
  end

  # @params new_headers is a [ [key, value], ...] array
  def add_headers!(*new_headers)
    new_headers.each { headers << _1 }
  end

  # @param headers_keys is a [word, word, ...] array
  def delete_headers!(*headers_keys)
    headers_keys.each { delete_header! _1 }
  end

  def delete_header!(header_key)
    headers.delete_if { _1[0].downcase == header_key.downcase }
  end

  def replace_headers!(*update_headers)
    delete_headers!(*update_headers.map(&:first))
    add_headers!(*update_headers)
  end

  private def smtp_headers
    headers.map { _1.join(": ") }.join("\r\n")
  end

  # TODO
  # List-Id: <mailto:>
  # List-Help: <mailto:>
  # List-Subscribe: <mailto:>
  # List-Unsubscribe: <mailto:>
  # List-Post: <mailto:>
  # List-Owner: <mailto:>
  # List-Archive: <https://...>
  # Archived-At: <https://...>
  # List-Post: NO (posting not allowed on this list)

  # https://www.ietf.org/rfc/rfc2369.txt
  # https://www.rfc-editor.org/rfc/rfc5983
  def to_smtp
    [
      "#{smtp_headers}\r\n\r\n#{@body}",
      @from,
      @to,
    ]
  end
end

==> v0.3.tar.gz: mailinglist.rb/lib/protocols/mail.rb <==
require "uuid"

class Protocols::Mail
  attr_accessor :from_name, :from, :to_name, :to, :subject, :body
  DATE_FORMAT = "%a, %d %b %Y %H:%M:%S %z".freeze

  HEADERS_KEEP = %w[Message-ID Subject From To Date In-Reply-To References Content-Type Content-Transfer-Encoding User-Agent MIME-Version]
  USER_AGENT = "Ruby/#{RUBY_VERSION} (#{RUBY_PLATFORM}) Mailinglist.rb/1.0"

  def initialize(imap_mail: nil)
    if imap_mail
      @imap_mail = imap_mail
      envelope = imap_mail.attr[Protocols::ENVELOPE]
      body_head = imap_mail.attr[Protocols::HEADERS]
      body_text = imap_mail.attr[Protocols::BODYTEXT]
      @from_name = envelope.from.first.name
      @to_name   = envelope.to.first.name
      @from      = "#{envelope.from.first.mailbox}@#{envelope.from.first.host}"
      @to        = "#{envelope.to.first.mailbox}@#{envelope.to.first.host}"
      @subject   = envelope.subject
      @body      = body_text
      @headers   = parse_rfc822_headers!(body_head)
      @seq       = header("X-Sequence").to_i
      @message_id = header("Message-id")
    end
  end

  def clone
    Protocols::Mail.new(imap_mail: @imap_mail)
  end

  def seen!(imap_client:)
    if @imap_mail
      imap_client.seen!(@imap_mail)
    else
      false
    end
  end

  # Transform an received email into a mailing list email.
  # Tested with evolution. Thunderbird bug at the moment.
  #
  # @param list   [Mailinglist]
  # @param dest [Email]
  def to_redistribute(list:, dest:)
    new = clone
    new.replace_headers!(
      ["User-Agent",       USER_AGENT],
      # require people to do not respond privatly
      ["Reply-to",         "#{list.name} <#{list.email}>"],
      ["Errors-To",        list.email],
      ["To",               dest.email],
      ["From",             "\"#{@from_name}\" (via #{list.name}) <#{@from}>"],
      ["Sender",           @from],
      ["Date",             Time.now.strftime(DATE_FORMAT)],
      ["List-Id",          "<#{list.email}>"],
      ["List-Post",        "<mailto:#{list.email}>"],
      ["List-Help",        "<mailto:#{list.help_email}>"],
      ["List-Subscribe",   "<mailto:#{list.subscribe_email}>"],
      ["List-Unsubscribe", "<mailto:#{list.unsubscribe_email}>"],
      ["List-Post",        "<mailto:#{list.post_email}>"],
      ["List-Owner",       "<mailto:#{list.owner_email}>"],
      ["In-Reply-To",      @message_id],
      ["Precedence",       "list"],
      # ["Precedence",       "bulk"],
      ["Message-Id",       "<#{UUID.generate}@#{ENV['SENDER_HOST']}>"],
      ["X-Sequence",         (@seq + 1).to_s],
      ["X-Loop",           list.email],
      ["Return-Path",      "<#{list.email}>"],
    )

    new.to   = dest.email
    new.from = list.email
    new.body = "#{body}#{list.signature}"

    new
  end

  # Create a email from scratch.
  #
  # @param list [Mailinglist]
  # @param to   [String]        email of the target
  # @param cc   [Array(String)] optional, emails (["mail@host", "Name <mail@host>", ...])
  def self.build(subject:, list:, to:, body:, cc: [])
    new = new()
    new.replace_headers!(
      ["User-Agent",       USER_AGENT],
      # require people to do not respond privatly
      ["Reply-to",         "#{list.name} <#{list.email}>"],
      ["Errors-To",        list.email],
      ["To",               to],
      ["Subject",          subject],
      ["From",             "\"#{list.name}\" <#{list.email}>"],
      ["Sender",           list.email],
      ["Date",             Time.now.strftime(DATE_FORMAT)],
      ["Message-Id",       "<#{UUID.generate}@#{ENV['SENDER_HOST']}>"],
    )
    new.replace_headers!(["CC", cc.join(", ")]) if !cc.empty?

    new.subject = subject
    new.to   = to
    new.from = list.email
    new.body = "#{body}#{list.signature}"
    new
  end

  # Create a email from scratch.
  # Create a response email for an existing one.
  #
  # @param list [Mailinglist]
  # @param to   [Protocols::Mail]
  def to_response(list:, to:, body:)
    new = clone
    new.replace_headers!(
      ["User-Agent",       USER_AGENT],
      # require people to do not respond privatly
      ["Reply-to",         "#{list.name} <#{list.email}>"],
      ["Errors-To",        list.email],
      ["To",               list.email],
      ["From",             "\"#{list.name}\" <#{list.email}>"],
      ["Sender",           list.email],
      ["Date",             Time.now.strftime(DATE_FORMAT)],
      ["Message-Id",       "<#{UUID.generate}@#{ENV['SENDER_HOST']}>"],
    )

    new.to   = to.from
    new.from = list.email
    new.body = "#{body}#{list.signature}"
    new
  end

  def parse_rfc822_headers!(rfc822)
    headers = []
    rfc822.split("\r\n").each do |line|
      if line[0] == "\t"
        headers.last[1] << "\r\n" << line
      else
        k, v = line.split(":", 2)
        headers << [k, v.to_s.strip]
      end
    end
    headers.filter! { HEADERS_KEEP.include?(_1[0]) }
  end

  def header(k)
    tuple = headers.find { _1[0].downcase == k.downcase }
    tuple && tuple[1]
  end

  def headers
    @headers ||= []
  end

  # @params new_headers is a [ [key, value], ...] array
  def add_headers!(*new_headers)
    new_headers.each { headers << _1 }
  end

  # @param headers_keys is a [word, word, ...] array
  def delete_headers!(*headers_keys)
    headers_keys.each { delete_header! _1 }
  end

  def delete_header!(header_key)
    headers.delete_if { _1[0].downcase == header_key.downcase }
  end

  def replace_headers!(*update_headers)
    delete_headers!(*update_headers.map(&:first))
    add_headers!(*update_headers)
  end

  private def smtp_headers
    headers.map { _1.join(": ") }.join("\r\n")
  end

  # TODO
  # List-Id: <mailto:>
  # List-Help: <mailto:>
  # List-Subscribe: <mailto:>
  # List-Unsubscribe: <mailto:>
  # List-Post: <mailto:>
  # List-Owner: <mailto:>
  # List-Archive: <https://...>
  # Archived-At: <https://...>
  # List-Post: NO (posting not allowed on this list)

  # https://www.ietf.org/rfc/rfc2369.txt
  # https://www.rfc-editor.org/rfc/rfc5983
  def to_smtp
    [
      "#{smtp_headers}\r\n\r\n#{@body}",
      @from,
      @to,
    ]
  end
end
